@@ -1,26 +1,6 @@
-/* PipeWire
- *
- * Copyright © 2018 Wim Taymans
- *
- * Permission is hereby granted, free of charge, to any person obtaining a
- * copy of this software and associated documentation files (the "Software"),
- * to deal in the Software without restriction, including without limitation
- * the rights to use, copy, modify, merge, publish, distribute, sublicense,
- * and/or sell copies of the Software, and to permit persons to whom the
- * Software is furnished to do so, subject to the following conditions:
- *
- * The above copyright notice and this permission notice (including the next
- * paragraph) shall be included in all copies or substantial portions of the
- * Software.
- *
- * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
- * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
- * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
- * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
- * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
- * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
- * DEALINGS IN THE SOFTWARE.
- */
+/* PipeWire */
+/* SPDX-FileCopyrightText: Copyright © 2018 Wim Taymans */
+/* SPDX-License-Identifier: MIT */
 
 #include "config.h"
 
@@ -35,24 +15,33 @@
 #include <fcntl.h>
 #include <sys/file.h>
 #include <ctype.h>
+#include <limits.h>
 #ifdef HAVE_PWD_H
 #include <pwd.h>
 #endif
+#ifdef HAVE_GRP_H
+#include <grp.h>
+#endif
 #if defined(__FreeBSD__) || defined(__MidnightBSD__)
 #include <sys/ucred.h>
 #endif
 
-#include <spa/pod/iter.h>
+#include <spa/pod/parser.h>
+#include <spa/pod/builder.h>
+#include <spa/utils/cleanup.h>
 #include <spa/utils/result.h>
 #include <spa/utils/string.h>
+#include <spa/utils/json.h>
+#include <spa/debug/log.h>
 
-#ifdef HAVE_SYSTEMD
-#include <systemd/sd-daemon.h>
+#ifdef HAVE_SELINUX
+#include <selinux/selinux.h>
 #endif
 
 #include <pipewire/impl.h>
 #include <pipewire/extensions/protocol-native.h>
 
+#include "network-utils.h"
 #include "pipewire/private.h"
 
 #include "modules/module-protocol-native/connection.h"
@@ -71,17 +60,36 @@ PW_LOG_TOPIC(mod_topic_connection, "conn." NAME);
 #include <spa/debug/pod.h>
 #include <spa/debug/types.h>
 
-/** \page page_module_protocol_native PipeWire Module: Protocol Native
+/** \page page_module_protocol_native Protocol Native
  *
  * The native protocol module implements the PipeWire communication between
  * a client and a server using unix local sockets.
  *
  * Normally this module is loaded in both client and server config files
- * so that they cam communicate.
+ * so that they can communicate.
+ *
+ * ## Module Name
+ *
+ * `libpipewire-module-protocol-native`
  *
  * ## Module Options
  *
- * The module has no options.
+ * The module supports the following arguments:
+ *
+ * - `sockets`: `[ { name = "socket-name", owner = "owner", group = "group", mode = "mode", selinux.context = "context" }, props = { ... }, ... ]`
+ *
+ *   Array of Unix socket names and (optionally) owner/permissions to serve,
+ *   if the context is a server. If not absolute paths, the sockets are created
+ *   in the default runtime directory.
+ *
+ *   The props are copied directly to any client that connects through this server
+ *   socket and can be used to configure special permissions.
+ *
+ *   Has the default value `[ { name = "CORENAME" }, { name = "CORENAME-manager" } ]`,
+ *   where `CORENAME` is the name of the PipeWire core, usually `pipewire-0`.
+ *
+ *   The permissions have no effect for sockets from Systemd socket activation.
+ *   Those should be configured via the systemd.socket(5) mechanism.
  *
  * ## General Options
  *
@@ -118,6 +126,22 @@ PW_LOG_TOPIC(mod_topic_connection, "conn." NAME);
  * local context. This can be done even when the server is not a daemon. It can
  * be used to treat a local context as if it was a server.
  *
+ * ## Config override
+ *
+ * A `module.protocol-native.args` config section can be added
+ * to override the module arguments.
+ *
+ *\code{.unparsed}
+ * # ~/.config/pipewire/pipewire.conf.d/my-protocol-native-args.conf
+ *
+ * module.protocol-native.args = {
+ *        sockets = [
+ *            { name = "pipewire-0" }
+ *            { name = "pipewire-0-manager" }
+ *        ]
+ * }
+ *\endcode
+ *
  * ## Example configuration
  *
  *\code{.unparsed}
@@ -125,6 +149,22 @@ PW_LOG_TOPIC(mod_topic_connection, "conn." NAME);
    { name = libpipewire-module-protocol-native }
  * ]
  *\endcode
+ *
+ *\code{.unparsed}
+ * context.modules = [
+ *  { name = libpipewire-module-protocol-native,
+ *    args = {
+ *        sockets = [
+ *            { name = "pipewire-0" }
+ *            { name = "pipewire-0-manager" }
+ *            { name = "pipewire-1"
+ *              props = { my.connection = "the other one" }
+ *            }
+ *        ]
+ *    }
+ *  }
+ * ]
+ *\endcode
  */
 
 #ifndef UNIX_PATH_MAX
@@ -142,19 +182,21 @@ static const struct spa_dict_item module_props[] = {
 #define SO_PEERSEC 31
 #endif
 
-static bool debug_messages = 0;
-
 #define LOCK_SUFFIX     ".lock"
 #define LOCK_SUFFIXLEN  5
 
 void pw_protocol_native_init(struct pw_protocol *protocol);
-void pw_protocol_native0_init(struct pw_protocol *protocol);
+void *protocol_native_security_context_init(struct pw_impl_module *module, struct pw_protocol *protocol);
+void protocol_native_security_context_free(void *data);
 
 struct protocol_data {
 	struct pw_impl_module *module;
 	struct spa_hook module_listener;
 	struct pw_protocol *protocol;
 
+	struct pw_properties *props;
+	void *security;
+
 	struct server *local;
 };
 
@@ -183,6 +225,16 @@ static void client_unref(struct client *impl)
 		free(impl);
 }
 
+struct socket_info {
+	char *name;
+	uid_t uid;
+	gid_t gid;
+	int mode;
+	char *selinux_context;
+	unsigned int has_owner:1;
+	unsigned int has_mode:1;
+};
+
 struct server {
 	struct pw_protocol_server this;
 
@@ -190,9 +242,12 @@ struct server {
 	struct sockaddr_un addr;
 	char lock_addr[UNIX_PATH_MAX + LOCK_SUFFIXLEN];
 
+	struct pw_properties *props;
+
 	struct pw_loop *loop;
 	struct spa_source *source;
 	struct spa_source *resume;
+	struct spa_source *close;
 	unsigned int activated:1;
 };
 
@@ -211,23 +266,29 @@ struct client_data {
 
 	unsigned int busy:1;
 	unsigned int need_flush:1;
-
-	struct protocol_compat_v2 compat_v2;
 };
 
 static void debug_msg(const char *prefix, const struct pw_protocol_native_message *msg, bool hex)
 {
 	struct spa_pod *pod;
+	struct spa_debug_log_ctx c;
+
+	if (!pw_log_topic_custom_enabled(SPA_LOG_LEVEL_DEBUG, mod_topic_connection))
+		return;
+
+	c = SPA_LOGT_DEBUG_INIT(pw_log_get(),
+				SPA_LOG_LEVEL_DEBUG, mod_topic_connection);
+
 	pw_logt_debug(mod_topic_connection,
-		      "%s: id:%d op:%d size:%d seq:%d", prefix,
-		      msg->id, msg->opcode, msg->size, msg->seq);
+		      "%s: id:%d op:%d size:%d seq:%d fds:%d", prefix,
+		      msg->id, msg->opcode, msg->size, msg->seq, msg->n_fds);
 
 	if ((pod = get_first_pod_from_data(msg->data, msg->size, 0)) != NULL)
-		spa_debug_pod(0, NULL, pod);
+		spa_debugc_pod(&c.ctx, 0, NULL, pod);
 	else
 		hex = true;
 	if (hex)
-		spa_debug_mem(0, msg->data, msg->size);
+		spa_debugc_mem(&c.ctx, 0, msg->data, msg->size);
 
 	pw_logt_debug(mod_topic_connection, "%s ****", prefix);
 
@@ -305,6 +366,7 @@ process_messages(struct client_data *data)
 			break;
 
 		if (client->core_resource == NULL) {
+			pw_log_debug("%p: no core resource", client);
 			res = -EPROTO;
 			goto error;
 		}
@@ -314,7 +376,7 @@ process_messages(struct client_data *data)
 		pw_log_trace("%p: got message %d from %u", client->protocol,
 			     msg->opcode, msg->id);
 
-		if (debug_messages)
+		if (pw_log_topic_custom_enabled(SPA_LOG_LEVEL_DEBUG, mod_topic_connection))
 			debug_msg("<<<<<< in", msg, false);
 
 		pre_demarshal(conn, msg, client, footer_client_demarshal,
@@ -467,8 +529,6 @@ static void client_free(void *data)
 		pw_loop_destroy_source(client->context->main_loop, this->source);
 	if (this->connection)
 		pw_protocol_native_connection_destroy(this->connection);
-
-	pw_map_clear(&this->compat_v2.types);
 }
 
 static const struct pw_impl_client_events client_events = {
@@ -498,9 +558,6 @@ static void on_start(void *data, uint32_t version)
 			PW_PERM_ALL, version, 0) < 0)
 		return;
 
-	if (version == 0)
-		client->compat_v2 = &this->compat_v2;
-
 	return;
 }
 
@@ -553,7 +610,7 @@ static struct client_data *client_new(struct server *s, int fd)
 	struct protocol_data *d = pw_protocol_get_user_data(protocol);
 	int i, res;
 
-	props = pw_properties_new(PW_KEY_PROTOCOL, "protocol-native", NULL);
+	props = pw_properties_copy(s->props);
 	if (props == NULL)
 		goto exit;
 
@@ -618,7 +675,6 @@ static struct client_data *client_new(struct server *s, int fd)
 
 	this->server = s;
 	this->client = client;
-	pw_map_init(&this->compat_v2.types, 0, 32);
 
 	pw_impl_client_add_listener(client, &this->client_listener, &client_events, this);
 
@@ -677,7 +733,7 @@ static int init_socket_name(struct server *s, const char *name)
 	const char *runtime_dir;
 	bool path_is_absolute;
 
-	path_is_absolute = name[0] == '/';
+	path_is_absolute = name[0] == '/' || name[0] == '@';
 
 	runtime_dir = get_runtime_dir();
 
@@ -715,6 +771,9 @@ static int lock_socket(struct server *s)
 {
 	int res;
 
+	if (s->addr.sun_path[0] == '\0')
+		return 0;
+
 	snprintf(s->lock_addr, sizeof(s->lock_addr), "%s%s", s->addr.sun_path, LOCK_SUFFIX);
 
 	s->fd_lock = open(s->lock_addr, O_CREAT | O_CLOEXEC,
@@ -768,6 +827,17 @@ socket_data(void *data, int fd, uint32_t mask)
 	}
 }
 
+static void
+close_data(void *data, int fd, uint32_t mask)
+{
+	struct server *s = data;
+
+	if (mask & (SPA_IO_HUP | SPA_IO_ERR)) {
+		pw_log_info("server %p: closed socket %d %08x", s, fd, mask);
+		pw_protocol_server_destroy(&s->this);
+	}
+}
+
 static int write_socket_address(struct server *s)
 {
 	long v;
@@ -806,19 +876,42 @@ error:
 	return res;
 }
 
-static int add_socket(struct pw_protocol *protocol, struct server *s)
+static int set_socket_permissions(struct server *s, struct socket_info *info)
+{
+	const char *path = s->addr.sun_path;
+
+	if (info->has_owner)
+		if (chown(path, info->uid, info->gid) < 0)
+			return -errno;
+
+	if (info->has_mode)
+		if (chmod(path, info->mode) < 0)
+			return -errno;
+
+	if (info->selinux_context) {
+#ifdef HAVE_SELINUX
+		if (setfilecon(path, info->selinux_context) < 0)
+			return -errno;
+#else
+		return -EOPNOTSUPP;
+#endif
+	}
+
+	return 0;
+}
+
+static int add_socket(struct pw_protocol *protocol, struct server *s, struct socket_info *info)
 {
 	socklen_t size;
 	int fd = -1, res;
 	bool activated = false;
 
-#ifdef HAVE_SYSTEMD
 	{
-		int i, n = sd_listen_fds(0);
+		int i, n = listen_fd();
 		for (i = 0; i < n; ++i) {
-			if (sd_is_socket_unix(SD_LISTEN_FDS_START + i, SOCK_STREAM,
-						1, s->addr.sun_path, 0) > 0) {
-				fd = SD_LISTEN_FDS_START + i;
+			if (is_socket_unix(LISTEN_FDS_START + i, SOCK_STREAM,
+						s->addr.sun_path) > 0) {
+				fd = LISTEN_FDS_START + i;
 				activated = true;
 				pw_log_info("server %p: Found socket activation socket for '%s'",
 						s, s->addr.sun_path);
@@ -826,7 +919,6 @@ static int add_socket(struct pw_protocol *protocol, struct server *s)
 			}
 		}
 	}
-#endif
 
 	if (fd < 0) {
 		struct stat socket_stat;
@@ -835,37 +927,48 @@ static int add_socket(struct pw_protocol *protocol, struct server *s)
 			res = -errno;
 			goto error;
 		}
-		if (stat(s->addr.sun_path, &socket_stat) < 0) {
-			if (errno != ENOENT) {
-				res = -errno;
-				pw_log_error("server %p: stat %s failed with error: %m",
-						s, s->addr.sun_path);
-				goto error_close;
+		if (s->addr.sun_path[0] == '@') {
+			s->addr.sun_path[0] = 0;
+			size = (socklen_t) (strlen(&s->addr.sun_path[1]) + 1);
+		} else {
+			if (stat(s->addr.sun_path, &socket_stat) < 0) {
+				if (errno != ENOENT) {
+					res = -errno;
+					pw_log_error("server %p: stat %s failed with error: %m",
+							s, s->addr.sun_path);
+					goto error_close;
+				}
+			} else if (socket_stat.st_mode & S_IWUSR || socket_stat.st_mode & S_IWGRP) {
+				unlink(s->addr.sun_path);
 			}
-		} else if (socket_stat.st_mode & S_IWUSR || socket_stat.st_mode & S_IWGRP) {
-			unlink(s->addr.sun_path);
+			size = (socklen_t) (strlen(s->addr.sun_path) + 1);
 		}
 
-		size = offsetof(struct sockaddr_un, sun_path) + strlen(s->addr.sun_path);
+		size += offsetof(struct sockaddr_un, sun_path);
 		if (bind(fd, (struct sockaddr *) &s->addr, size) < 0) {
 			res = -errno;
 			pw_log_error("server %p: bind() failed with error: %m", s);
 			goto error_close;
 		}
 
+		if ((res = set_socket_permissions(s, info)) < 0) {
+			errno = -res;
+			pw_log_error("server %p: failed to set socket %s permissions: %m",
+					s, info->name);
+			goto error_close;
+		}
+
 		if (listen(fd, 128) < 0) {
 			res = -errno;
 			pw_log_error("server %p: listen() failed with error: %m", s);
 			goto error_close;
 		}
+	} else {
+		if (info->has_owner || info->has_mode || info->selinux_context)
+			pw_log_info("server %p: permissions ignored for socket %s from systemd",
+					s, info->name);
 	}
 
-	res = write_socket_address(s);
-	if (res < 0) {
-		pw_log_error("server %p: failed to write socket address: %s", s,
-				spa_strerror(res));
-		goto error_close;
-	}
 	s->activated = activated;
 	s->loop = pw_context_get_main_loop(protocol->context);
 	if (s->loop == NULL) {
@@ -877,6 +980,11 @@ static int add_socket(struct pw_protocol *protocol, struct server *s)
 		res = -errno;
 		goto error_close;
 	}
+	res = write_socket_address(s);
+	if (res < 0) {
+		pw_log_warn("server %p: failed to write socket address: %s", s,
+				spa_strerror(res));
+	}
 	return 0;
 
 error_close:
@@ -930,7 +1038,7 @@ process_remote(struct client *impl)
 
 		this->recv_seq = msg->seq;
 
-		if (debug_messages)
+		if (pw_log_topic_custom_enabled(SPA_LOG_LEVEL_DEBUG, mod_topic_connection))
 			debug_msg("<<<<<< in", msg, false);
 
 		pre_demarshal(conn, msg, this, footer_core_demarshal,
@@ -938,12 +1046,18 @@ process_remote(struct client *impl)
 
 		proxy = pw_core_find_proxy(this, msg->id);
 		if (proxy == NULL || proxy->zombie) {
+			uint32_t i;
+
 			if (proxy == NULL)
 				pw_log_error("%p: could not find proxy %u", this, msg->id);
 			else
 				pw_log_debug("%p: zombie proxy %u", this, msg->id);
 
-			/* FIXME close fds */
+			/* close fds */
+			for (i = 0; i < msg->n_fds; i++) {
+				pw_log_debug("%p: close fd:%d", conn, msg->fds[i]);
+				close(msg->fds[i]);
+			}
 			continue;
 		}
 
@@ -1206,9 +1320,8 @@ impl_new_client(struct pw_protocol *protocol,
 
 	if (props) {
 		str = spa_dict_lookup(props, PW_KEY_REMOTE_INTENTION);
-		if (str == NULL &&
-		   (str = spa_dict_lookup(props, PW_KEY_REMOTE_NAME)) != NULL &&
-		    spa_streq(str, "internal"))
+		if ((str == NULL || spa_streq(str, "generic")) &&
+		   spa_streq(spa_dict_lookup(props, PW_KEY_REMOTE_NAME), "internal"))
 			str = "internal";
 	}
 	if (str == NULL)
@@ -1255,12 +1368,15 @@ static void destroy_server(struct pw_protocol_server *server)
 		pw_loop_destroy_source(s->loop, s->source);
 	if (s->resume)
 		pw_loop_destroy_source(s->loop, s->resume);
+	if (s->close)
+		pw_loop_destroy_source(s->loop, s->close);
 	if (s->addr.sun_path[0] && !s->activated)
 		unlink(s->addr.sun_path);
 	if (s->lock_addr[0])
 		unlink(s->lock_addr);
 	if (s->fd_lock != -1)
 		close(s->fd_lock);
+	pw_properties_free(s->props);
 	free(s);
 }
 
@@ -1287,10 +1403,9 @@ get_server_name(const struct spa_dict *props)
 {
 	const char *name = NULL;
 
-	if (props)
+	name = getenv("PIPEWIRE_CORE");
+	if (name == NULL && props != NULL)
 		name = spa_dict_lookup(props, PW_KEY_CORE_NAME);
-	if (name == NULL)
-		name = getenv("PIPEWIRE_CORE");
 	if (name == NULL)
 		name = PW_DEFAULT_REMOTE;
 	return name;
@@ -1307,6 +1422,12 @@ create_server(struct pw_protocol *protocol,
 	if ((s = calloc(1, sizeof(struct server))) == NULL)
 		return NULL;
 
+	s->props = pw_properties_new(PW_KEY_PROTOCOL, "protocol-native", NULL);
+	if (s->props == NULL) {
+		free(s);
+		return NULL;
+	}
+	pw_properties_update(s->props, props);
 	s->fd_lock = -1;
 
 	this = &s->this;
@@ -1314,7 +1435,6 @@ create_server(struct pw_protocol *protocol,
 	this->core = core;
 	spa_list_init(&this->client_list);
 	this->destroy = destroy_server;
-
 	spa_list_append(&protocol->server_list, &this->link);
 
 	pw_log_debug("%p: created server %p", protocol, this);
@@ -1323,9 +1443,10 @@ create_server(struct pw_protocol *protocol,
 }
 
 static struct pw_protocol_server *
-impl_add_server(struct pw_protocol *protocol,
+add_server(struct pw_protocol *protocol,
 		struct pw_impl_core *core,
-                const struct spa_dict *props)
+		const struct spa_dict *props,
+		struct socket_info *socket_info)
 {
 	struct pw_protocol_server *this;
 	struct server *s;
@@ -1337,7 +1458,12 @@ impl_add_server(struct pw_protocol *protocol,
 
 	this = &s->this;
 
-	name = get_server_name(props);
+	if (socket_info)
+		name = socket_info->name;
+	else
+		name = get_server_name(props);
+
+	pw_properties_set(s->props, PW_KEY_SEC_SOCKET, name);
 
 	if ((res = init_socket_name(s, name)) < 0)
 		goto error;
@@ -1345,11 +1471,13 @@ impl_add_server(struct pw_protocol *protocol,
 	if ((res = lock_socket(s)) < 0)
 		goto error;
 
-	if ((res = add_socket(protocol, s)) < 0)
+	if ((res = add_socket(protocol, s, socket_info)) < 0)
 		goto error;
 
-	if ((s->resume = pw_loop_add_event(s->loop, do_resume, s)) == NULL)
+	if ((s->resume = pw_loop_add_event(s->loop, do_resume, s)) == NULL) {
+		res = -errno;
 		goto error;
+	}
 
 	pw_log_info("%p: Listening on '%s'", protocol, name);
 
@@ -1361,10 +1489,66 @@ error:
 	return NULL;
 }
 
+static struct pw_protocol_server *
+impl_add_server(struct pw_protocol *protocol,
+		struct pw_impl_core *core,
+                const struct spa_dict *props)
+{
+	return add_server(protocol, core, props, NULL);
+}
+
+static struct pw_protocol_server *
+impl_add_fd_server(struct pw_protocol *protocol,
+		struct pw_impl_core *core,
+		int listen_fd, int close_fd,
+                const struct spa_dict *props)
+{
+	struct pw_protocol_server *this;
+	struct server *s;
+	int res;
+
+	if ((s = create_server(protocol, core, props)) == NULL)
+		return NULL;
+
+	this = &s->this;
+
+	pw_properties_setf(s->props, PW_KEY_SEC_SOCKET, "pipewire-fd-%d", listen_fd);
+
+	s->loop = pw_context_get_main_loop(protocol->context);
+	if (s->loop == NULL) {
+		res = -errno;
+		goto error;
+	}
+	s->source = pw_loop_add_io(s->loop, listen_fd, SPA_IO_IN, true, socket_data, s);
+	if (s->source == NULL) {
+		res = -errno;
+		goto error;
+	}
+	s->close = pw_loop_add_io(s->loop, close_fd, 0, true, close_data, s);
+	if (s->close == NULL) {
+		res = -errno;
+		goto error;
+	}
+	if ((s->resume = pw_loop_add_event(s->loop, do_resume, s)) == NULL) {
+		res = -errno;
+		goto error;
+	}
+
+	pw_log_info("%p: Listening on fd:%d", protocol, listen_fd);
+
+	return this;
+
+error:
+	destroy_server(this);
+	errno = -res;
+	return NULL;
+}
+
 static const struct pw_protocol_implementation protocol_impl = {
 	PW_VERSION_PROTOCOL_IMPLEMENTATION,
 	.new_client = impl_new_client,
 	.add_server = impl_add_server,
+	.add_fd_server = impl_add_fd_server,
 };
 
 static struct spa_pod_builder *
@@ -1403,6 +1587,7 @@ static int impl_ext_end_proxy(struct pw_proxy *proxy,
 {
 	struct pw_core *core = proxy->core;
 	struct client *impl = SPA_CONTAINER_OF(core->conn, struct client, this);
+	ensure_loop(impl->context->main_loop);
 	assert_single_pod(builder);
 	marshal_core_footers(&impl->footer_state, core, builder);
 	return core->send_seq = pw_protocol_native_connection_end(impl->connection, builder);
@@ -1432,6 +1617,7 @@ static int impl_ext_end_resource(struct pw_resource *resource,
 {
 	struct client_data *data = resource->client->user_data;
 	struct pw_impl_client *client = resource->client;
+	ensure_loop(client->context->main_loop);
 	assert_single_pod(builder);
 	marshal_client_footers(&data->footer_state, client, builder);
 	return client->send_seq = pw_protocol_native_connection_end(data->connection, builder);
@@ -1452,7 +1638,10 @@ static void module_destroy(void *data)
 {
 	struct protocol_data *d = data;
 
+	if (d->security)
+		protocol_native_security_context_free(d->security);
 	spa_hook_remove(&d->module_listener);
+	pw_properties_free(d->props);
 
 	pw_protocol_destroy(d->protocol);
 }
@@ -1466,57 +1655,198 @@ static int need_server(struct pw_context *context, const struct spa_dict *props)
 {
 	const char *val = NULL;
 
-	if (props)
+	val = getenv("PIPEWIRE_DAEMON");
+	if (val == NULL && props != NULL)
 		val = spa_dict_lookup(props, PW_KEY_CORE_DAEMON);
-	if (val == NULL)
-		val = getenv("PIPEWIRE_DAEMON");
 	if (val && pw_properties_parse_bool(val))
 		return 1;
 	return 0;
 }
 
+static int create_servers(struct pw_protocol *this, struct pw_impl_core *core,
+		const struct pw_properties *props, const struct pw_properties *args)
+{
+	const char *sockets = args ? pw_properties_get(args, "sockets") : NULL;
+	struct spa_json it[2];
+	spa_autoptr(pw_properties) p = pw_properties_copy(props);
+
+	if (sockets == NULL) {
+		struct socket_info info = {0};
+		spa_autofree char *manager_name = NULL;
+
+		info.name = (char *)get_server_name(&p->dict);
+		if (add_server(this, core, &p->dict, &info) == NULL)
+			return -errno;
+
+		manager_name = spa_aprintf("%s-manager", info.name);
+		if (manager_name == NULL)
+			return -ENOMEM;
+
+		info.name = manager_name;
+		if (add_server(this, core, &p->dict, &info) == NULL)
+			return -errno;
+
+		return 0;
+	}
+
+	if (spa_json_begin_array(&it[0], sockets, strlen(sockets)) <= 0)
+		goto error_invalid;
+
+	while (spa_json_enter_object(&it[0], &it[1]) > 0) {
+		struct socket_info info = {0};
+		char key[256];
+		char name[PATH_MAX];
+		char selinux_context[PATH_MAX];
+		const char *value;
+		int len;
+
+		info.uid = getuid();
+		info.gid = getgid();
+
+		pw_properties_clear(p);
+		pw_properties_update(p, &props->dict);
+
+		while ((len = spa_json_object_next(&it[1], key, sizeof(key), &value)) > 0) {
+			if (spa_streq(key, "name")) {
+				if (spa_json_parse_stringn(value, len, name, sizeof(name)) < 0)
+					goto error_invalid;
+				info.name = name;
+			} else if (spa_streq(key, "selinux.context")) {
+				if (spa_json_parse_stringn(value, len, selinux_context, sizeof(selinux_context)) < 0)
+					goto error_invalid;
+				info.selinux_context = selinux_context;
+			} else if (spa_streq(key, "owner")) {
+				char buffer[16384];
+				char owner[PATH_MAX];
+				struct passwd pwd, *result = NULL;
+				int64_t val;
+
+				if (spa_json_parse_stringn(value, len, owner, sizeof(owner)) < 0)
+					goto error_invalid;
+
+				if (spa_atoi64(owner, &val, 10))
+					info.uid = val;
+				else if (getpwnam_r(owner, &pwd, buffer, sizeof(buffer), &result) == 0 && result)
+					info.uid = result->pw_uid;
+				else
+					goto error_invalid;
+
+				info.has_owner = true;
+			} else if (spa_streq(key, "group")) {
+				char buffer[16384];
+				char group[PATH_MAX];
+				struct group grp, *result = NULL;
+				int64_t val;
+
+				if (spa_json_parse_stringn(value, len, group, sizeof(group)) < 0)
+					goto error_invalid;
+
+				if (spa_atoi64(group, &val, 10))
+					info.gid = val;
+				else if (getgrnam_r(group, &grp, buffer, sizeof(buffer), &result) == 0 && result)
+					info.gid = result->gr_gid;
+				else
+					goto error_invalid;
+
+				info.has_owner = true;
+			} else if (spa_streq(key, "mode")) {
+				char mode[PATH_MAX];
+				int64_t val;
+
+				if (spa_json_parse_stringn(value, len, mode, sizeof(mode)) < 0)
+					goto error_invalid;
+
+				if (spa_atoi64(mode, &val, 0))
+					info.mode = val;
+				else
+					goto error_invalid;
+
+				info.has_mode = true;
+			} else if (spa_streq(key, "props")) {
+				if (spa_json_is_container(value, len))
+	                                len = spa_json_container_len(&it[1], value, len);
+
+				pw_properties_update_string(p, value, len);
+			}
+		}
+
+		if (info.name == NULL)
+			goto error_invalid;
+
+		if (add_server(this, core, &p->dict, &info) == NULL)
+			return -errno;
+	}
+
+	return 0;
+
+error_invalid:
+	pw_log_error("invalid module 'sockets' argument: %s", sockets);
+	return -EINVAL;
+}
+
 SPA_EXPORT
-int pipewire__module_init(struct pw_impl_module *module, const char *args)
+int pipewire__module_init(struct pw_impl_module *module, const char *args_str)
 {
 	struct pw_context *context = pw_impl_module_get_context(module);
 	struct pw_protocol *this;
+	struct pw_impl_core *core = context->core;
 	struct protocol_data *d;
 	const struct pw_properties *props;
+	spa_autoptr(pw_properties) args = NULL;
 	int res;
+	static const char * const keys[] = {
+		PW_KEY_CORE_NAME,
+		NULL
+	};
 
 	PW_LOG_TOPIC_INIT(mod_topic);
 	PW_LOG_TOPIC_INIT(mod_topic_connection);
 
-	if (pw_context_find_protocol(context, PW_TYPE_INFO_PROTOCOL_Native) != NULL)
-		return 0;
+	if (pw_context_find_protocol(context, PW_TYPE_INFO_PROTOCOL_Native) != NULL) {
+		pw_log_error("protocol %s is already loaded", PW_TYPE_INFO_PROTOCOL_Native);
+		return -EEXIST;
+	}
+
+	args = args_str ? pw_properties_new_string(args_str) : pw_properties_new(NULL, NULL);
+	if (!args)
+		return -errno;
+
+	pw_context_conf_update_props(context, "module."NAME".args", args);
 
 	this = pw_protocol_new(context, PW_TYPE_INFO_PROTOCOL_Native, sizeof(struct protocol_data));
 	if (this == NULL)
 		return -errno;
 
-	debug_messages = mod_topic_connection->level >= SPA_LOG_LEVEL_DEBUG;
-
 	this->implementation = &protocol_impl;
 	this->extension = &protocol_ext_impl;
 
 	pw_protocol_native_init(this);
-	pw_protocol_native0_init(this);
 
-	pw_log_debug("%p: new debug:%d", this, debug_messages);
+	pw_log_debug("%p: new", this);
 
 	d = pw_protocol_get_user_data(this);
 	d->protocol = this;
 	d->module = module;
+	d->props = pw_properties_new(NULL, NULL);
+	if (d->props == NULL) {
+		res = -ENOMEM;
+		goto error_cleanup;
+	}
+
+	d->security = protocol_native_security_context_init(module, this);
+	if (d->security == NULL) {
+		res = -errno;
+		goto error_cleanup;
+	}
 
 	props = pw_context_get_properties(context);
-	d->local = create_server(this, context->core, &props->dict);
+	pw_properties_update_keys(d->props, &props->dict, keys);
 
-	if (need_server(context, &props->dict)) {
-		if (impl_add_server(this, context->core, &props->dict) == NULL) {
-			res = -errno;
+	d->local = create_server(this, core, &d->props->dict);
+
+	if (need_server(context, &props->dict))
+		if ((res = create_servers(this, core, d->props, args)) < 0)
 			goto error_cleanup;
-		}
-	}
 
 	pw_impl_module_add_listener(module, &d->module_listener, &module_events, d);
 
@@ -1525,6 +1855,9 @@ int pipewire__module_init(struct pw_impl_module *module, const char *args)
 	return 0;
 
 error_cleanup:
+	pw_properties_free(d->props);
+	if (d->security)
+		protocol_native_security_context_free(d->security);
 	pw_protocol_destroy(this);
 	return res;
 }
