@@ -1,26 +1,6 @@
-/* PipeWire
- *
- * Copyright © 2020 Wim Taymans
- *
- * Permission is hereby granted, free of charge, to any person obtaining a
- * copy of this software and associated documentation files (the "Software"),
- * to deal in the Software without restriction, including without limitation
- * the rights to use, copy, modify, merge, publish, distribute, sublicense,
- * and/or sell copies of the Software, and to permit persons to whom the
- * Software is furnished to do so, subject to the following conditions:
- *
- * The above copyright notice and this permission notice (including the next
- * paragraph) shall be included in all copies or substantial portions of the
- * Software.
- *
- * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
- * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
- * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
- * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
- * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
- * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
- * DEALINGS IN THE SOFTWARE.
- */
+/* PipeWire */
+/* SPDX-FileCopyrightText: Copyright © 2020 Wim Taymans */
+/* SPDX-License-Identifier: MIT */
 
 #include "config.h"
 
@@ -41,15 +21,14 @@
 #include <netinet/ip.h>
 #include <unistd.h>
 
-#ifdef HAVE_SYSTEMD
-#include <systemd/sd-daemon.h>
-#endif
 
+#include <spa/utils/cleanup.h>
 #include <spa/utils/defs.h>
 #include <spa/utils/json.h>
 #include <spa/utils/result.h>
 #include <pipewire/pipewire.h>
 
+#include "network-utils.h"
 #include "client.h"
 #include "commands.h"
 #include "defs.h"
@@ -61,10 +40,15 @@
 #include "stream.h"
 #include "utils.h"
 #include "flatpak-utils.h"
+#ifdef HAVE_SNAP
+#include "snap-policy.h"
+#endif
 
 #define LISTEN_BACKLOG 32
 #define MAX_CLIENTS 64
 
+PW_LOG_TOPIC_EXTERN(pulse_conn);
+
 static int handle_packet(struct client *client, struct message *msg)
 {
 	uint32_t command, tag;
@@ -86,9 +70,9 @@ static int handle_packet(struct client *client, struct message *msg)
 		goto finish;
 	}
 
-	if (debug_messages) {
+	if (pw_log_topic_custom_enabled(SPA_LOG_LEVEL_INFO, pulse_conn)) {
 		pw_log_debug("client %p: command:%s", client, commands[command].name);
-		message_dump(SPA_LOG_LEVEL_INFO, msg);
+		message_dump(SPA_LOG_LEVEL_INFO, "<<", msg);
 	}
 
 	const struct command *cmd = &commands[command];
@@ -117,6 +101,15 @@ finish:
 	return 0;
 }
 
+static void stream_clear_data(struct stream *stream,
+		uint32_t offset, uint32_t len)
+{
+	uint32_t l0 = SPA_MIN(len, MAXLENGTH - offset), l1 = len - l0;
+	sample_spec_silence(&stream->ss, SPA_PTROFF(stream->buffer, offset, void), l0);
+	if (SPA_UNLIKELY(l1 > 0))
+		sample_spec_silence(&stream->ss, stream->buffer, l1);
+}
+
 static int handle_memblock(struct client *client, struct message *msg)
 {
 	struct stream *stream;
@@ -163,6 +156,15 @@ static int handle_memblock(struct client *client, struct message *msg)
 		goto finish;
 	}
 
+	if (diff > 0) {
+		pw_log_debug("clear gap of %"PRIu64, diff);
+		/* if we jump forwards, clear the data we skipped because we might otherwise
+		 * play back old data. FIXME, if the write pointer goes backwards and
+		 * forwards, this might clear valid data. We should probably keep track of
+		 * the highest write pointer and only clear when we go past that one. */
+		stream_clear_data(stream, index % MAXLENGTH, SPA_MIN(diff, MAXLENGTH));
+	}
+
 	index += diff;
 	filled += diff;
 	stream->write_index += diff;
@@ -370,6 +372,8 @@ on_connect(void *data, int fd, uint32_t mask)
 	int client_fd, val;
 	struct client *client = NULL;
 	const char *client_access = NULL;
+	const char *error_reason = NULL;
+	char ipname[256];
 	pid_t pid;
 
 	length = sizeof(name);
@@ -379,7 +383,7 @@ on_connect(void *data, int fd, uint32_t mask)
 			if (server->n_clients > 0) {
 				int m = server->source->mask;
 				SPA_FLAG_CLEAR(m, SPA_IO_IN);
-				pw_loop_update_io(impl->loop, server->source, m);
+				pw_loop_update_io(impl->main_loop, server->source, m);
 				server->wait_clients++;
 			}
 		}
@@ -388,6 +392,7 @@ on_connect(void *data, int fd, uint32_t mask)
 
 	if (server->n_clients >= server->max_clients) {
 		close(client_fd);
+		error_reason = "too many client application connections";
 		errno = ECONNREFUSED;
 		goto error;
 	}
@@ -398,7 +403,7 @@ on_connect(void *data, int fd, uint32_t mask)
 
 	pw_log_debug("server %p: new client %p fd:%d", server, client, client_fd);
 
-	client->source = pw_loop_add_io(impl->loop,
+	client->source = pw_loop_add_io(impl->main_loop,
 					client_fd,
 					SPA_IO_ERR | SPA_IO_HUP | SPA_IO_IN,
 					true, on_client_data, client);
@@ -412,10 +417,18 @@ on_connect(void *data, int fd, uint32_t mask)
 	if (client->props == NULL)
 		goto error;
 
+
 	pw_properties_setf(client->props,
 			"pulse.server.type", "%s",
 			server->addr.ss_family == AF_UNIX ? "unix" : "tcp");
 
+	if (server->addr.ss_family != AF_UNIX) {
+		uint16_t port = 0;
+		if (pw_net_get_ip(&name, ipname, sizeof(ipname), NULL, &port) >= 0)
+			pw_properties_setf(client->props,
+					"pulse.server.peer", "%s:%d", ipname, port);
+	}
+
 	client->routes = pw_properties_new(NULL, NULL);
 	if (client->routes == NULL)
 		goto error;
@@ -424,7 +437,10 @@ on_connect(void *data, int fd, uint32_t mask)
 		client_access = server->client_access;
 
 	if (server->addr.ss_family == AF_UNIX) {
-		char *app_id = NULL, *devices = NULL;
+		spa_autofree char *app_id = NULL, *snap_app_id = NULL, *devices = NULL, *instance_id = NULL;
+#ifdef HAVE_SNAP
+		pw_sandbox_access_t snap_access;
+#endif
 
 #ifdef SO_PRIORITY
 		val = 6;
@@ -432,7 +448,7 @@ on_connect(void *data, int fd, uint32_t mask)
 			pw_log_warn("setsockopt(SO_PRIORITY) failed: %m");
 #endif
 		pid = get_client_pid(client, client_fd);
-		if (pid != 0 && pw_check_flatpak(pid, &app_id, &devices) == 1) {
+		if (pid != 0 && pw_check_flatpak(pid, &app_id, &instance_id, &devices) == 1) {
 			/*
 			 * XXX: we should really use Portal client access here
 			 *
@@ -455,6 +471,8 @@ on_connect(void *data, int fd, uint32_t mask)
 			client_access = "flatpak";
 			pw_properties_set(client->props, "pipewire.access.portal.app_id",
 					app_id);
+			pw_properties_set(client->props, "pipewire.access.portal.instance_id",
+					instance_id);
 
 			if (devices && (spa_streq(devices, "all") ||
 							spa_strstartswith(devices, "all;") ||
@@ -463,8 +481,21 @@ on_connect(void *data, int fd, uint32_t mask)
 			else
 				pw_properties_set(client->props, PW_KEY_MEDIA_CATEGORY, NULL);
 		}
-		free(devices);
-		free(app_id);
+		// check SNAP permissions
+#ifdef HAVE_SNAP
+		snap_access = pw_snap_get_audio_permissions(client, client_fd, &snap_app_id);
+		if ((snap_access & PW_SANDBOX_ACCESS_NOT_A_SANDBOX) == 0) {
+			pw_properties_set(client->props, PW_KEY_SNAP_ID, snap_app_id);
+
+			pw_properties_set(client->props,
+					  PW_KEY_SNAP_PLAYBACK_ALLOWED,
+					  (snap_access & PW_SANDBOX_ACCESS_PLAYBACK) ? "true" : "false");
+
+			pw_properties_set(client->props,
+					  PW_KEY_SNAP_RECORD_ALLOWED,
+					  (snap_access & PW_SANDBOX_ACCESS_RECORD) ? "true" : "false");
+		}
+#endif
 	}
 	else if (server->addr.ss_family == AF_INET || server->addr.ss_family == AF_INET6) {
 
@@ -485,7 +516,9 @@ on_connect(void *data, int fd, uint32_t mask)
 	return;
 
 error:
-	pw_log_error("server %p: failed to create client: %m", server);
+	pw_log_error("server %p: %s: %m", server,
+			error_reason ? error_reason : "failed to create client");
+
 	if (client)
 		client_free(client);
 }
@@ -541,26 +574,19 @@ static bool is_stale_socket(int fd, const struct sockaddr_un *addr_un)
 	return false;
 }
 
-#ifdef HAVE_SYSTEMD
-static int check_systemd_activation(const char *path)
+static int check_socket_activation(const char *path)
 {
-	const int n = sd_listen_fds(0);
+	const int n = listen_fd();
 
 	for (int i = 0; i < n; i++) {
-		const int fd = SD_LISTEN_FDS_START + i;
+		const int fd = LISTEN_FDS_START + i;
 
-		if (sd_is_socket_unix(fd, SOCK_STREAM, 1, path, 0) > 0)
+		if (is_socket_unix(fd, SOCK_STREAM, path) > 0)
 			return fd;
 	}
 
 	return -1;
 }
-#else
-static inline int check_systemd_activation(SPA_UNUSED const char *path)
-{
-	return -1;
-}
-#endif
 
 static int start_unix_server(struct server *server, const struct sockaddr_storage *addr)
 {
@@ -570,10 +596,10 @@ static int start_unix_server(struct server *server, const struct sockaddr_storag
 
 	spa_assert(addr_un->sun_family == AF_UNIX);
 
-	fd = check_systemd_activation(addr_un->sun_path);
+	fd = check_socket_activation(addr_un->sun_path);
 	if (fd >= 0) {
 		server->activated = true;
-		pw_log_info("server %p: found systemd socket activation socket for '%s'",
+		pw_log_info("server %p: found socket activation socket for '%s'",
 			    server, addr_un->sun_path);
 		goto done;
 	}
@@ -823,13 +849,14 @@ static int parse_ip_address(const char *address, struct sockaddr_storage *addrs,
 	if (len < 2)
 		return -ENOSPC;
 
-	snprintf(ip, sizeof(ip), "[::]:%d", res);
-	spa_assert_se(parse_ipv6_address(ip, (struct sockaddr_in6 *) &addr) == 0);
-	addrs[0] = addr;
-
 	snprintf(ip, sizeof(ip), "0.0.0.0:%d", res);
 	spa_assert_se(parse_ipv4_address(ip, (struct sockaddr_in *) &addr) == 0);
+	addrs[0] = addr;
+
+	snprintf(ip, sizeof(ip), "[::]:%d", res);
+	spa_assert_se(parse_ipv6_address(ip, (struct sockaddr_in6 *) &addr) == 0);
 	addrs[1] = addr;
+
 	return 2;
 }
 
@@ -922,7 +949,7 @@ static int server_start(struct server *server, const struct sockaddr_storage *ad
 	if (fd < 0)
 		return fd;
 
-	server->source = pw_loop_add_io(impl->loop, fd, SPA_IO_IN, true, on_connect, server);
+	server->source = pw_loop_add_io(impl->main_loop, fd, SPA_IO_IN, true, on_connect, server);
 	if (server->source == NULL) {
 		res = -errno;
 		pw_log_error("server %p: can't create server source: %m", impl);
@@ -974,31 +1001,27 @@ int servers_create_and_start(struct impl *impl, const char *addresses, struct pw
 {
 	int len, res, count = 0, err = 0; /* store the first error to return when no servers could be created */
 	const char *v;
-	struct spa_json it[3];
+	struct spa_json it[2];
 
 	/* update `err` if it hasn't been set to an errno */
 #define UPDATE_ERR(e) do { if (err == 0) err = (e); } while (false)
 
 	/* collect addresses into an array of `struct sockaddr_storage` */
-	spa_json_init(&it[0], addresses, strlen(addresses));
 
 	/* [ <server-spec> ... ] */
-	if (spa_json_enter_array(&it[0], &it[1]) < 0)
+	if (spa_json_begin_array(&it[0], addresses, strlen(addresses)) < 0)
 		return -EINVAL;
 
 	/* a server-spec is either an address or an object */
-	while ((len = spa_json_next(&it[1], &v)) > 0) {
+	while ((len = spa_json_next(&it[0], &v)) > 0) {
 		char addr_str[FORMATTED_SOCKET_ADDR_STRLEN] = { 0 };
 		char key[128], client_access[64] = { 0 };
 		struct sockaddr_storage addrs[2];
 		int i, max_clients = MAX_CLIENTS, listen_backlog = LISTEN_BACKLOG, n_addr;
 
 		if (spa_json_is_object(v, len)) {
-			spa_json_enter(&it[1], &it[2]);
-			while (spa_json_get_string(&it[2], key, sizeof(key)) > 0) {
-				if ((len = spa_json_next(&it[2], &v)) <= 0)
-					break;
-
+			spa_json_enter(&it[0], &it[1]);
+			while ((len = spa_json_object_next(&it[1], key, sizeof(key), &v)) > 0) {
 				if (spa_streq(key, "address")) {
 					spa_json_parse_stringn(v, len, addr_str, sizeof(addr_str));
 				} else if (spa_streq(key, "max-clients")) {
@@ -1077,7 +1100,7 @@ void server_free(struct server *server)
 	spa_hook_list_call(&impl->hooks, struct impl_events, server_stopped, 0, server);
 
 	if (server->source)
-		pw_loop_destroy_source(impl->loop, server->source);
+		pw_loop_destroy_source(impl->main_loop, server->source);
 
 	if (server->addr.ss_family == AF_UNIX && !server->activated)
 		unlink(((const struct sockaddr_un *) &server->addr)->sun_path);
