@@ -3,15 +3,16 @@
 /* SPDX-FileCopyrightText: Copyright © 2021 Sanchayan Maity <sanchayan@asymptotic.io> */
 /* SPDX-License-Identifier: MIT */
 
+#include "config.h"
+
 #include <limits.h>
 #include <sys/stat.h>
 #include <unistd.h>
 
-#include "config.h"
-
 #include <spa/utils/hook.h>
 #include <spa/utils/result.h>
 #include <spa/param/audio/format-utils.h>
+#include <spa/param/audio/raw-json.h>
 
 #include <roc/config.h>
 #include <roc/log.h>
@@ -24,12 +25,16 @@
 
 #include "module-roc/common.h"
 
-/** \page page_module_roc_source PipeWire Module: ROC source
+/** \page page_module_roc_source ROC source
  *
  * The `roc-source` module creates a PipeWire source that receives samples
  * from ROC sender and passes them to the sink it is connected to. One can
  * then connect it to any audio device.
  *
+ * ## Module Name
+ *
+ * `libpipewire-module-roc-source`
+ *
  * ## Module Options
  *
  * Options specific to the behavior of this module
@@ -39,10 +44,21 @@
  * - `local.ip = <str>`: local sender ip
  * - `local.source.port = <str>`: local receiver TCP/UDP port for source packets
  * - `local.repair.port = <str>`: local receiver TCP/UDP port for receiver packets
+ * - `local.control.port = <str>`: local receiver TCP/UDP port for control packets
  * - `sess.latency.msec = <str>`: target network latency in milliseconds
- * - `resampler.profile = <str>`: Possible values: `disable`, `high`,
- *   `medium`, `low`.
- * - `fec.code = <str>`: Possible values: `disable`, `rs8m`, `ldpc`
+ * - `roc.resampler.backend = <str>`: Possible values: `default`, `builtin`,
+ *       `speex`, `speexdec`.
+ * - `roc.resampler.profile = <str>`: Possible values: `default`, `high`,
+ *       `medium`, `low`.
+ * - `roc.latency-tuner.backend = <str>`: Possible values: `default`, `niq`
+ * - `roc.latency-tuner.profile = <str>`: Possible values: `default`, `intact`,
+ *       `responsive`, `gradual`
+ * - `fec.code = <str>`: Possible values: `default`, `disable`, `rs8m`, `ldpc`
+ *
+ * - `resampler.profile = <str>`: Deprecated, use roc.resampler.profile
+ * - `log.level = <str>`: log level for roc-toolkit. Possible values: `DEFAULT`,
+ *       `NONE`, `ERROR`, `INFO`, `DEBUG`, `TRACE`; `DEFAULT` follows the log
+ * level of the PipeWire context.
  *
  * ## General options
  *
@@ -50,23 +66,33 @@
  *
  * - \ref PW_KEY_NODE_NAME
  * - \ref PW_KEY_NODE_DESCRIPTION
- * - \ref PW_KEY_MEDIA_NAME
+ * - \ref PW_KEY_NODE_VIRTUAL
+ * - \ref PW_KEY_MEDIA_CLASS
+ * - \ref SPA_KEY_AUDIO_POSITION
  *
  * ## Example configuration
  *\code{.unparsed}
+ * # ~/.config/pipewire/pipewire.conf.d/my-roc-source.conf
+ *
  * context.modules = [
  *  {   name = libpipewire-module-roc-source
  *      args = {
  *          local.ip = 0.0.0.0
- *          resampler.profile = medium
+ *          #roc.resampler.backend = default
+ *          roc.resampler.profile = medium
+ *          #roc.latency-tuner.backend = default
+ *          #roc.latency-tuner.profile = default
  *          fec.code = disable
  *          sess.latency.msec = 5000
  *          local.source.port = 10001
  *          local.repair.port = 10002
+ *          local.control.port = 10003
  *          source.name = "ROC Source"
  *          source.props = {
  *             node.name = "roc-source"
  *          }
+ *          audio.position = [ FL FR ]
+ *          log.level = DEFAULT
  *      }
  *  }
  *]
@@ -76,13 +102,13 @@
 
 #define NAME "roc-source"
 
-PW_LOG_TOPIC_STATIC(mod_topic, "mod." NAME);
+PW_LOG_TOPIC(mod_topic, "mod." NAME);
 #define PW_LOG_TOPIC_DEFAULT mod_topic
+PW_LOG_TOPIC_EXTERN(roc_log_topic);
 
 struct module_roc_source_data {
 	struct pw_impl_module *module;
 	struct spa_hook module_listener;
-	struct pw_properties *props;
 	struct pw_context *module_context;
 
 	struct pw_core *core;
@@ -102,12 +128,20 @@ struct module_roc_source_data {
 	roc_receiver *receiver;
 
 	roc_resampler_profile resampler_profile;
+	roc_resampler_backend resampler_backend;
+	roc_latency_tuner_backend latency_tuner_backend;
+	roc_latency_tuner_profile latency_tuner_profile;
 	roc_fec_encoding fec_code;
 	uint32_t rate;
 	char *local_ip;
 	int local_source_port;
 	int local_repair_port;
 	int sess_latency_msec;
+
+	roc_endpoint *local_control_addr;
+	int local_control_port;
+
+	roc_log_level loglevel;
 };
 
 static void stream_destroy(void *d)
@@ -216,17 +250,13 @@ static void impl_destroy(struct module_roc_source_data *data)
 		pw_core_disconnect(data->core);
 
 	pw_properties_free(data->playback_props);
-	pw_properties_free(data->props);
 
-	if (data->receiver)
-		roc_receiver_close(data->receiver);
-	if (data->context)
-		roc_context_close(data->context);
+	spa_clear_ptr(data->receiver, roc_receiver_close);
+	spa_clear_ptr(data->context, roc_context_close);
 
-	if (data->local_source_addr)
-		(void) roc_endpoint_deallocate(data->local_source_addr);
-	if (data->local_repair_addr)
-		(void) roc_endpoint_deallocate(data->local_repair_addr);
+	spa_clear_ptr(data->local_source_addr, roc_endpoint_deallocate);
+	spa_clear_ptr(data->local_repair_addr, roc_endpoint_deallocate);
+	spa_clear_ptr(data->local_control_addr, roc_endpoint_deallocate);
 
 	free(data->local_ip);
 	free(data);
@@ -264,18 +294,35 @@ static int roc_source_setup(struct module_roc_source_data *data)
 	}
 
 	spa_zero(receiver_config);
-	receiver_config.frame_sample_rate = data->rate;
-	receiver_config.frame_channels = ROC_CHANNEL_SET_STEREO;
-	receiver_config.frame_encoding = ROC_FRAME_ENCODING_PCM_FLOAT;
-	receiver_config.resampler_profile = data->resampler_profile;
 
-	info.rate = data->rate;
+	receiver_config.frame_encoding.rate = data->rate;
+	receiver_config.frame_encoding.format = ROC_FORMAT_PCM_FLOAT32;
+	receiver_config.resampler_profile = data->resampler_profile;
+	receiver_config.resampler_backend = data->resampler_backend;
+	receiver_config.latency_tuner_backend = data->latency_tuner_backend;
+	receiver_config.latency_tuner_profile = data->latency_tuner_profile;
+
 
 	/* Fixed to be the same as ROC receiver config above */
-	info.channels = 2;
+	info.rate = data->rate;
 	info.format = SPA_AUDIO_FORMAT_F32;
-	info.position[0] = SPA_AUDIO_CHANNEL_FL;
-	info.position[1] = SPA_AUDIO_CHANNEL_FR;
+
+	const char* positions = pw_properties_get(data->playback_props, SPA_KEY_AUDIO_POSITION);
+	int channels = spa_audio_parse_position_n(positions, strlen(positions), info.position, SPA_N_ELEMENTS(info.position), &info.channels);
+
+	if(channels == 2) {
+		receiver_config.frame_encoding.channels = ROC_CHANNEL_LAYOUT_STEREO;
+	} else {
+		receiver_config.frame_encoding.channels = ROC_CHANNEL_LAYOUT_MULTITRACK;
+		receiver_config.frame_encoding.tracks = channels;
+
+		res = roc_context_register_encoding(data->context, PW_ROC_MULTITRACK_ENCODING_ID, &receiver_config.frame_encoding);
+		if(res) {
+			pw_log_error("failed to register encoding: %d", res);
+			return -EINVAL;
+		}
+	}
+
 	data->stride = info.channels * sizeof(float);
 
 	pw_properties_setf(data->playback_props, PW_KEY_NODE_RATE, "1/%d", info.rate);
@@ -291,7 +338,9 @@ static int roc_source_setup(struct module_roc_source_data *data)
 	 * See API reference:
 	 * https://roc-streaming.org/toolkit/docs/api/reference.html
 	 */
-	receiver_config.target_latency = data->sess_latency_msec * 1000000;
+	receiver_config.target_latency = (unsigned long long)data->sess_latency_msec * SPA_NSEC_PER_MSEC;
+
+	pw_roc_log_init();
 
 	res = roc_receiver_open(data->context, &receiver_config, &data->receiver);
 	if (res) {
@@ -299,21 +348,7 @@ static int roc_source_setup(struct module_roc_source_data *data)
 		return -EINVAL;
 	}
 
-	switch (data->fec_code) {
-	case ROC_FEC_ENCODING_DEFAULT:
-	case ROC_FEC_ENCODING_RS8M:
-		audio_proto = ROC_PROTO_RTP_RS8M_SOURCE;
-		repair_proto = ROC_PROTO_RS8M_REPAIR;
-		break;
-	case ROC_FEC_ENCODING_LDPC_STAIRCASE:
-		audio_proto = ROC_PROTO_RTP_LDPC_SOURCE;
-		repair_proto = ROC_PROTO_LDPC_REPAIR;
-		break;
-	default:
-		audio_proto = ROC_PROTO_RTP;
-		repair_proto = 0;
-		break;
-	}
+	pw_roc_fec_encoding_to_proto(data->fec_code, &audio_proto, &repair_proto);
 
 	res = pw_roc_create_endpoint(&data->local_source_addr, audio_proto, data->local_ip, data->local_source_port);
 	if (res < 0) {
@@ -341,6 +376,18 @@ static int roc_source_setup(struct module_roc_source_data *data)
 		}
 	}
 
+	res = pw_roc_create_endpoint(&data->local_control_addr, PW_ROC_DEFAULT_CONTROL_PROTO, data->local_ip, data->local_control_port);
+	if (res < 0) {
+		pw_log_error("failed to create control endpoint: %s", spa_strerror(res));
+		return res;
+	}
+
+	if (roc_receiver_bind(data->receiver, ROC_SLOT_DEFAULT, ROC_INTERFACE_AUDIO_CONTROL,
+				data->local_control_addr) != 0) {
+		pw_log_error("can't connect roc receiver to local control address");
+		return -EINVAL;
+	}
+
 	data->playback = pw_stream_new(data->core,
 			"roc-source playback", data->playback_props);
 	data->playback_props = NULL;
@@ -372,13 +419,19 @@ static const struct spa_dict_item module_roc_source_info[] = {
 	{ PW_KEY_MODULE_AUTHOR, "Sanchayan Maity <sanchayan@asymptotic.io>" },
 	{ PW_KEY_MODULE_DESCRIPTION, "roc source" },
 	{ PW_KEY_MODULE_USAGE,	"( source.name=<name for the source> ) "
-				"( resampler.profile=<empty>|disable|high|medium|low ) "
+				"( roc.resampler.backend=<empty>|default|builtin|speex|speexdec ) "
+				"( roc.resampler.profile=<empty>|default|high|medium|low ) "
+				"( roc.latency-tuner.backend=<empty>|default|niq ) "
+				"( roc.latency-tuner.profile=<empty>|default|intact|responsive|gradual ) "
 				"( fec.code=<empty>|disable|rs8m|ldpc ) "
 				"( sess.latency.msec=<target network latency in milliseconds> ) "
 				"( local.ip=<local receiver ip> ) "
 				"( local.source.port=<local receiver port for source packets> ) "
 				"( local.repair.port=<local receiver port for repair packets> ) "
-				"( source.props= { key=value ... } ) " },
+				"( local.control.port=<local receiver port for control packets> ) "
+				"( audio.position=<channel map, default:"PW_ROC_STEREO_POSITIONS"> ) "
+				"( source.props= { key=value ... } ) "
+				"( log.level=<empty>|DEFAULT|NONE|RROR|INFO|DEBUG|TRACE ) " },
 	{ PW_KEY_MODULE_VERSION, PACKAGE_VERSION },
 };
 
@@ -387,7 +440,7 @@ int pipewire__module_init(struct pw_impl_module *module, const char *args)
 {
 	struct pw_context *context = pw_impl_module_get_context(module);
 	struct module_roc_source_data *data;
-	struct pw_properties *props = NULL, *playback_props = NULL;
+	struct pw_properties *playback_props = NULL;
 	const char *str;
 	int res = 0;
 
@@ -400,13 +453,12 @@ int pipewire__module_init(struct pw_impl_module *module, const char *args)
 	if (args == NULL)
 		args = "";
 
-	props = pw_properties_new_string(args);
+	spa_autoptr(pw_properties) props = pw_properties_new_string(args);
 	if (props == NULL) {
 		res = -errno;
 		pw_log_error( "can't create properties: %m");
 		goto out;
 	}
-	data->props = props;
 
 	playback_props = pw_properties_new(NULL, NULL);
 	if (playback_props == NULL) {
@@ -421,7 +473,6 @@ int pipewire__module_init(struct pw_impl_module *module, const char *args)
 
 	if ((str = pw_properties_get(props, "source.name")) != NULL) {
 		pw_properties_set(playback_props, PW_KEY_NODE_NAME, str);
-		pw_properties_set(props, "source.name", NULL);
 	}
 
 	if ((str = pw_properties_get(props, "source.props")) != NULL)
@@ -436,56 +487,94 @@ int pipewire__module_init(struct pw_impl_module *module, const char *args)
 	if (pw_properties_get(playback_props, PW_KEY_NODE_NETWORK) == NULL)
 		pw_properties_set(playback_props, PW_KEY_NODE_NETWORK, "true");
 
-	data->rate = pw_properties_get_uint32(playback_props, PW_KEY_AUDIO_RATE, data->rate);
+	if ((str = pw_properties_get(props, SPA_KEY_AUDIO_POSITION)) != NULL) {
+		pw_properties_set(playback_props, SPA_KEY_AUDIO_POSITION, str);
+	} else {
+		pw_properties_set(playback_props, SPA_KEY_AUDIO_POSITION, PW_ROC_STEREO_POSITIONS);
+	}
+
+	data->rate = pw_properties_get_uint32(playback_props, PW_KEY_AUDIO_RATE, 0);
 	if (data->rate == 0)
 		data->rate = PW_ROC_DEFAULT_RATE;
 
 	if ((str = pw_properties_get(props, "local.ip")) != NULL) {
 		data->local_ip = strdup(str);
-		pw_properties_set(props, "local.ip", NULL);
 	} else {
 		data->local_ip = strdup(PW_ROC_DEFAULT_IP);
 	}
 
 	if ((str = pw_properties_get(props, "local.source.port")) != NULL) {
 		data->local_source_port = pw_properties_parse_int(str);
-		pw_properties_set(props, "local.source.port", NULL);
 	} else {
 		data->local_source_port = PW_ROC_DEFAULT_SOURCE_PORT;
 	}
 
 	if ((str = pw_properties_get(props, "local.repair.port")) != NULL) {
 		data->local_repair_port = pw_properties_parse_int(str);
-		pw_properties_set(props, "local.repair.port", NULL);
 	} else {
 		data->local_repair_port = PW_ROC_DEFAULT_REPAIR_PORT;
 	}
 
+	if ((str = pw_properties_get(props, "local.control.port")) != NULL) {
+		data->local_control_port = pw_properties_parse_int(str);
+	} else {
+		data->local_control_port = PW_ROC_DEFAULT_CONTROL_PORT;
+	}
+
 	if ((str = pw_properties_get(props, "sess.latency.msec")) != NULL) {
 		data->sess_latency_msec = pw_properties_parse_int(str);
-		pw_properties_set(props, "sess.latency.msec", NULL);
 	} else {
 		data->sess_latency_msec = PW_ROC_DEFAULT_SESS_LATENCY;
 	}
-
-	if ((str = pw_properties_get(props, "resampler.profile")) != NULL) {
+	if ((str = pw_properties_get(props, "roc.resampler.backend")) != NULL) {
+		if (pw_roc_parse_resampler_backend(&data->resampler_backend, str)) {
+			pw_log_warn("Invalid resampler backend %s, using default", str);
+			data->resampler_backend = ROC_RESAMPLER_BACKEND_DEFAULT;
+		}
+	} else {
+		data->resampler_backend = ROC_RESAMPLER_BACKEND_DEFAULT;
+	}
+	if ((str = pw_properties_get(props, "roc.resampler.profile")) != NULL ||
+	    (str = pw_properties_get(props, "resampler.profile")) != NULL) {
 		if (pw_roc_parse_resampler_profile(&data->resampler_profile, str)) {
 			pw_log_warn("Invalid resampler profile %s, using default", str);
 			data->resampler_profile = ROC_RESAMPLER_PROFILE_DEFAULT;
 		}
-		pw_properties_set(props, "resampler.profile", NULL);
 	} else {
 		data->resampler_profile = ROC_RESAMPLER_PROFILE_DEFAULT;
 	}
+	if ((str = pw_properties_get(props, "roc.latency-tuner.backend")) != NULL) {
+		if (pw_roc_parse_latency_tuner_backend(&data->latency_tuner_backend, str)) {
+			pw_log_warn("Invalid latency-tuner backend %s, using default", str);
+			data->latency_tuner_backend = ROC_LATENCY_TUNER_BACKEND_DEFAULT;
+		}
+	} else {
+		data->latency_tuner_backend = ROC_LATENCY_TUNER_BACKEND_DEFAULT;
+	}
+	if ((str = pw_properties_get(props, "roc.latency-tuner.profile")) != NULL) {
+		if (pw_roc_parse_latency_tuner_profile(&data->latency_tuner_profile, str)) {
+			pw_log_warn("Invalid latency-tuner profile %s, using default", str);
+			data->latency_tuner_profile = ROC_LATENCY_TUNER_PROFILE_DEFAULT;
+		}
+	} else {
+		data->latency_tuner_profile = ROC_LATENCY_TUNER_PROFILE_DEFAULT;
+	}
 	if ((str = pw_properties_get(props, "fec.code")) != NULL) {
 		if (pw_roc_parse_fec_encoding(&data->fec_code, str)) {
 			pw_log_error("Invalid fec code %s, using default", str);
 			data->fec_code = ROC_FEC_ENCODING_DEFAULT;
 		}
-		pw_properties_set(props, "fec.code", NULL);
 	} else {
 		data->fec_code = ROC_FEC_ENCODING_DEFAULT;
 	}
+	if ((str = pw_properties_get(props, "log.level")) != NULL) {
+		const struct spa_log *log_conf = pw_log_get();
+		const roc_log_level default_level = pw_roc_log_level_pw_2_roc(log_conf->level);
+		if (pw_roc_parse_log_level(&data->loglevel, str, default_level)) {
+			pw_log_error("Invalid log level %s, using default", str);
+			data->loglevel = default_level;
+		}
+	}
 
 	data->core = pw_context_get_object(data->module_context, PW_TYPE_INTERFACE_Core);
 	if (data->core == NULL) {
